 /*
 * Trieda zobrazuje cislovanie riadkov
 */
package textArea;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;

import javax.swing.border.Border;

public class LeftWindow extends JComponent implements SwingConstants {

    public LeftWindow(WorkWindow textArea, TextAreaDefaults defaults) {
        this.textArea = textArea;

        setBackground(defaults.leftWindowBgColor);
        setForeground(defaults.leftWindowFgColor);
        setFont(defaults.leftWindowFont);
        setBorder(defaults.leftWindowBorderWidth, defaults.leftWindowBorderColor);
        setleftWindowWidth(defaults.leftWindowWidth);
        setCollapsed(defaults.leftWindowCollapsed);
        leftWindowMouseListener ml = new leftWindowMouseListener();
        addMouseListener(ml);
        addMouseMotionListener(ml);
    }

    public void paintComponent(Graphics gfx) {
        if (!collapsed) {
  
            Rectangle r = gfx.getClipBounds();
            gfx.setColor(getBackground());
            gfx.fillRect(r.x, r.y, r.width, r.height);
            paintLineNumbers(gfx);

        }
    }

    protected void paintLineNumbers(Graphics gfx) {
        FontMetrics pfm = textArea.getPainter().getFontMetrics();
        int lineHeight = pfm.getHeight();
        int baseline = (int) Math.round((this.baseline + lineHeight- pfm.getMaxDescent()) / 2.0);

        int firstLine = textArea.getFirstLine() + 1;
        int lastLine = firstLine + (getHeight() / lineHeight);

        int firstValidLine =  Math.max(1, firstLine);
        int lastValidLine =  Math.min(textArea.getLineCount(), lastLine);

        gfx.setFont(getFont());

        gfx.setColor(getForeground());

        String number;

        for (int line = firstLine; line <= lastLine; line++, baseline += lineHeight) {

            if (line < firstValidLine || line > lastValidLine) {
                continue;
            }

            number = Integer.toString(line);
            int offset = leftWindowSize.width - collapsedSize.width - (fm.stringWidth(number) + 1);

            if (interval > 1 && line % interval == 0) {

                gfx.drawString(number, ileft + offset, baseline);
                gfx.setColor(getForeground());
            } else {
                gfx.drawString(number, ileft + offset, baseline);
            }
        }
    }

    public void setBorder(int width, Color color) {
        setBorder(BorderFactory.createMatteBorder(0, 0, 0, width, color));
    }

    public void setBorder(Border border) {
        super.setBorder(border);

        if (border == null) {
            ileft = 0;
            collapsedSize.width = 0;
            collapsedSize.height = 0;
        } else {
            Insets insets = border.getBorderInsets(this);
            ileft = insets.left;
            collapsedSize.width = insets.left + insets.right;
            collapsedSize.height = insets.top + insets.bottom;
        }
    }

    public void setFont(Font font) {
        super.setFont(font);

        fm = getFontMetrics(font);
        baseline = fm.getHeight() - fm.getMaxDescent();
    }

    public void setleftWindowWidth(int width) {
        if (width < collapsedSize.width) {
            width = collapsedSize.width;
        }

        leftWindowSize.width = width;

             if (!collapsed) {
            textArea.revalidate();
        }
    }

    public int getleftWindowWidth() {
        return leftWindowSize.width;
    }

    public Dimension getPreferredSize() {
        if (collapsed) {
            return collapsedSize;
        } else {
            return leftWindowSize;
        }
    }

    public Dimension getMinimumSize() {
        return getPreferredSize();
    }

    public int getLineNumberAlignment() {
        return alignment;
    }

    public void setLineNumberAlignment(int alignment) {
        if (this.alignment == alignment) {
            return;
        }

        this.alignment = alignment;

        repaint();
    }

    public boolean isCollapsed() {
        return collapsed;
    }

    public void setCollapsed(boolean collapsed) {
        if (this.collapsed == collapsed) {
            return;
        }

        this.collapsed = collapsed;

        textArea.revalidate();
    }

    /**
     * Toggles whether the leftWindow is collapsed or expanded.
     */
    public void toggleCollapsed() {
        setCollapsed(!collapsed);
    }

    public int yToLine(int y) {
        FontMetrics pfm = textArea.getPainter().getFontMetrics();
        int lineHeight = pfm.getHeight();
        int baseline = (int) Math.round((this.baseline + lineHeight - pfm.getMaxDescent()) / 2.0);

        int firstLine = textArea.getFirstLine() + 1;
        int lastLine = firstLine + (getHeight() / lineHeight);

        int firstValidLine = Math.max(1, firstLine);
        int lastValidLine =  Math.min(textArea.getLineCount(), lastLine);
        return Math.max(0, Math.min(lastLine - 1,  y / lineHeight + firstLine));

    }
  
    private WorkWindow textArea;
    private int baseline = 0;
    private int ileft = 0;
    private Dimension leftWindowSize = new Dimension(0, 0);
    private Dimension collapsedSize = new Dimension(0, 0);
    private FontMetrics fm;
    private int alignment;
    private int interval = 0;
    private boolean collapsed = false;

    class leftWindowMouseListener extends MouseAdapter  implements MouseMotionListener {

        @Override
        public void mouseClicked(MouseEvent e) {
            int count = e.getClickCount();
            if (count == 1) {
                try {
                    int line = yToLine(e.getY()) - 1;

                    textArea.setCaretPosition(textArea.getLineStartOffset(line));

                } catch (Exception ee) {
                }
            } else if (count >= 2) {
                toggleCollapsed();
            }
        }

        public void mousePressed(MouseEvent e) {
            dragStart = e.getPoint();

            startWidth = leftWindowSize.width;
        }

        public void mouseDragged(MouseEvent e) {
            if (dragStart == null) {
                return;
            }

            if (isCollapsed()) {
                setCollapsed(false);
            }

            Point p = e.getPoint();

            leftWindowSize.width = startWidth + p.x - dragStart.x;

            if (leftWindowSize.width < collapsedSize.width) {
                leftWindowSize.width = startWidth;
                setCollapsed(true);
            }

            SwingUtilities.invokeLater(new Runnable() {

                public void run() {
                    textArea.revalidate();
                }
            });
        }

        public void mouseExited(MouseEvent e) {
            if (dragStart != null && dragStart.x > e.getPoint().x) {
                setCollapsed(true);
                leftWindowSize.width = startWidth;

                SwingUtilities.invokeLater(new Runnable() {

                    public void run() {
                        textArea.revalidate();
                    }
                });
            }

            //dragStart = null;
        }

       

        public void mouseReleased(MouseEvent e) {
            dragStart = null;
        }
        private Point dragStart = null;
        private int startWidth = 10;
    }
}
